

%%%%%%%%	Some Options	%%%%%%%%


%%	bpmn_atts_names	%%
%
%	Option flag. BPMN entities describing attributes
%	are named according to names specified in BPMN diagram.
%	This option has no effect on sequence flows attributes names.
%	See no_bpmn_atts_names/0.
%
bpmn_atts_names:-
	insert(is_bpmn_atts_names).

%%	no_bpmn_atts_names	%%
%
%	Option flag. BPMN entities describing attributes
%	names are generated. Oposite to bpmn_atts_names/0.
%
%	|	*BPMN* *type*	|	*Attribute* *name*		|
%	|	activity		|	a<number>				|
%	|	event			|	e<number>				|
%	|	sequence flow	|	f<number>				|
%
no_bpmn_atts_names:-
	delete(is_bpmn_atts_names).
:-dynamic is_bpmn_atts_names/0.

%%	callbacks_for_existential_exprs_atts_only	%%
%
%	Callbacks are generated only for
%	existential expressions.
%
%	Existential expression is in form: <attribute_name>
%	example BPMN diagram may look this way:
%
%	some_task ---- result ---- > other_task
%
%	In this case "result" denotes attribute name and it
%	will be translated to existential expression
%	see expr_existential/1. The "result" is called here as "existential attribute".
%
callbacks_for_existential_exprs_atts_only:-
	insert(is_callback_only_for_existential_exprs).
callbacks_for_each_exprs_atts:-
	delete(is_callback_only_for_existential_exprs).
:-dynamic is_callback_only_for_existential_exprs/0.




%%%%%%%%	Core predicates	%%%%%%%%


%%	bpmn_to_xtt(+BpmnFile,+XttFile)	%%
%
%	Translates BPMN located in BpmnFile (Eclipse STP format)
%	to XTT written to XttFile.
%
bpmn_to_xtt(BpmnFile,XttFile):-
	bpmn_to_xttelems(BpmnFile,AttsTypes,Atts,Xtts),
	xttelems_write(XttFile,AttsTypes,Atts,Xtts).


%%	xttelems_write(+XttFile,+AttsTypes,+Atts,+Xtts)
%
%	Writes XTT BPMN diagram representation
%	to XttFile see bpmn_to_xttelems/4.
%
xttelems_write(XttFile,AttsTypes,Atts,Xtts):-
	open(XttFile,write,Stream),

	write(Stream,'%%%%%%%%%%%%%%	Attributes Types	%%%%%%%%%%%%%%%%'),
	stream_nl(Stream),stream_nl(Stream),
	forall(member(TypeAtt,AttsTypes),(typeatt_write(TypeAtt,Stream),stream_nl(Stream))),

	stream_nl(Stream),stream_nl(Stream),

	write(Stream,'%%%%%%%%%%%%%%	Attributes	%%%%%%%%%%%%%%%%'),
	stream_nl(Stream),stream_nl(Stream),
	atts_write(Atts,Stream),

	stream_nl(Stream),stream_nl(Stream),

	write(Stream,'%%%%%%%%%%%%%%	XTT Tables	%%%%%%%%%%%%%%%%'),
	stream_nl(Stream),stream_nl(Stream),
	xtts_write(Xtts,Stream),

	stream_nl(Stream),stream_nl(Stream),
	write(Stream, 'xstat s1:[].'),stream_nl(Stream),

	stream_nl(Stream),stream_nl(Stream),
	stream_nl(Stream),stream_nl(Stream),

	write(Stream,'%%%%%%%%%%%%%%	Callbacks Stubs	%%%%%%%%%%%%%%%%'),
	stream_nl(Stream),stream_nl(Stream),

	write(Stream,'%%%%	Rules Callbacks	%%%%'),
	stream_nl(Stream),stream_nl(Stream),
	xtts_rules_callbacks_write(Xtts,Stream),

	stream_nl(Stream),stream_nl(Stream),

	write(Stream, '%%%%	Attributes Callbacks	%%%%'),
	stream_nl(Stream),stream_nl(Stream),
	atts_callbacks_write(Atts,Stream),
	close(Stream).


%%	bpmn_to_xttelems(+BpmnFile,-AttsTypes,-Attributes,-Xtts)
%
%	Read BpmnFile and translate to XTT elements.
%
%	@param	AttsTypes	List of XTT attributes types representations.
%	@param	Attributes	List of XTT attributes representations that describes BPMN diagram elements.
%	@param	Xtts		List of XTT tables representations that describes BPMN diagram logic.
%
bpmn_to_xttelems(BpmnFile,AttsTypes,Attributes,Xtts):-
	node_clear,att_clear,

	load_xml_file(BpmnFile,Xml),
	member(element('bpmn:BpmnDiagram', _, Pools), Xml),!,
	member(element(pools, _, ObjectsXml), Pools),

	node_actvs_all(ObjectsXml,Actvs),

	%% Divide to Ends and Rest
	node_find_all(ObjectsXml,'vertices',['activityType'='EventEndEmpty'],Ends),
	select_all(Ends,Actvs,Ends,ActvsR),

	%% rest activities processing
	actvs_to_xtts(ObjectsXml,ActvsR,AttsTypes,Atts0,Xtts1),


	% Ends processing
	actvs_to_onextt(ObjectsXml,Ends,AttsEnd,EndXtts),


	% postprocessing
	append(Xtts1,EndXtts,Xtts),
	append(Atts0,AttsEnd,Atts1),

	flatten(Atts1,Atts2),
	list_to_set(Atts2,Atts3),
	remove_empty(Atts3,Atts4),
	atts_finalize(Atts4,Attributes).




%%	actvs_to_xtts(+ObjectsXML,+Actvs,-AttsTypes,-Attss,-Xtts)	%%
%
%	High level transformation predicate
%	transforms Actvs (BPMN entities) in context of ObjectsXML
%	to corresponding XTT description.
%
%	@param	ObjectsXML	context of Actvs that are being translated (content of BPMN Pool that contains all Actvs)
%	@param	Actvs		list of XML nodes that represents BPMN entities
%	declared in same BPMN pool (without BPMN sequence flows (edges)).
%
%	@param	AttsTypes	XTT types definition (types of attributes).
%	@param	Attss		List of lists of XTT attributes representations.
%	@param	Xtts		List of XTT tables representations.
%
actvs_to_xtts(ObjectsXML,Actvs,AttsTypes,Attss,Xtts):-
	actvs_to_xtts1(ObjectsXML,Actvs,AttsTypes,Attss,Xtts).

actvs_to_xtts1(_,[],[],[],[]):-!.
actvs_to_xtts1(Xml,[Actv|Actvs],AttsTypes,[Atts|Attss],[Xtt|Xtts]):-

	actv_preprocess(Xml,Actv,Xmlobj,InEdges,OutEdges,_,_,EdgesAtts),

	actv_to_xtt(Xmlobj,InEdges,OutEdges,Xtt,Att),
	Atts = [Att|EdgesAtts],

	actvs_to_xtts1(Xml,Actvs,_,Attss,Xtts),

	% this is not effective but for refactoring purpose it is good step
	findall(AttType,typeatt_get(AttType),AttsTypes).

actvs_to_onextt(_,[],[],[]):-!.
actvs_to_onextt(Xml,Actvs,Atts,[Xtt]):-
	actvs_merge(Xml,Actvs,Xmlobj,InEdges,OutEdges,EdgesAtts),
	actv_to_xtt(Xmlobj,InEdges,OutEdges,Xtt,Att),
	Atts = [Att|EdgesAtts].


actvs_merge(Xml,Actvs,Xmlobj,InEdges,OutEdges,Atts):-
	actvs_merge1(Xml,Actvs,[],InEdges,[],OutEdges,[],Atts),
	member(Actv,Actvs),!,
	xmlobj_create(Xml,Actv,Xmlobj).

actvs_merge1(_,[],InEdges,InEdges,OutEdges,OutEdges,Atts,Atts):-!.
actvs_merge1(Xml,[Actv|Actvs],InEdges0,InEdges,OutEdges0,OutEdges,Atts0,Atts):-
	actv_preprocess(Xml,Actv,_,InEdges1,OutEdges1,_,_,Atts1),
	append(InEdges0,InEdges1,InEdges2),
	append(OutEdges0,OutEdges1,OutEdges2),
	append(Atts0,Atts1,Atts2),
	actvs_merge1(Xml,Actvs,InEdges2,InEdges,OutEdges2,OutEdges,Atts2,Atts).


actv_preprocess(Xml,Actv,Xmlobj,InEdges,OutEdges,InEdgesAtts,OutEdgesAtts,EdgesAtts):-
	xmlobj_create(Xml,Actv,Xmlobj),

	xmlobj_edges(Xmlobj,in,InEdges),
	xmlobj_edges(Xmlobj,out,OutEdges),

	edges_atts_and_conds_atts(InEdges,InEdgesAtts),
	edges_atts_and_conds_atts(OutEdges,OutEdgesAtts),
	append(InEdgesAtts,OutEdgesAtts,EdgesAtts).

actv_to_xtt(Xmlobj,InEdges,OutEdges,Xtt,Att):-
	xmlobj_type(Xmlobj,Type),
	actv_to_xtt1(Xmlobj,Type,InEdges,OutEdges,Xtt,Att).

%	Activities and Events transformation
actv_to_xtt1(Xmlobj,Type,InEdges,OutEdges,Xtt,Att):-
	(actv_type(Type),!;event_type(Type)),!,
	xtt_create(Schema,Rules,Xtt),

	edges_atts(InEdges,InAtts1),
	edges_conds_arg1_atts(InEdges,InAtts2),
	append(InAtts1,InAtts2,InAtts),

	edges_atts(OutEdges,OutAtts),
	append(OutAtts,[Att],Outs),

	att_name(Att,AttName),
	callback_no(NoCallback),
	xmlobj_to_att(Xmlobj,NoCallback,Att),

	get_xtttable_name(Xmlobj,XttName),
	schema_create(XttName,InAtts,Outs,Schema),


	callback_gen_name1(action,AttName,RuleCallbackName),
	callback_create(action,AttName,RuleCallbackName,ActvCallback),

	%connditions of InEdges are not handled

	%att_to_expristrue(Att,AttIsTrue),
	%rule_create(XttName,_,[AttIsTrue],[],[ActvCallback],Rule),

	rules_create_in_out(or,and,XttName,InEdges,OutEdges,Att,[ActvCallback],Rules).



%	Gateways transformation
actv_to_xtt1(Xmlobj,GatewayType,InEdges,OutEdges,Xtt,[]):-
	gateway_type(_,_,GatewayType),
	gateway_to_xtt(Xmlobj,GatewayType,InEdges,OutEdges,Xtt).

gateway_to_xtt(Xmlobj,GatewayType,InEdges,OutEdges,Xtt):-!,
	gateway_sj_type(InEdges,OutEdges,SplitOrJoin),!,
	gateway_type(_,SplitOrJoin,GatewayType),
	xtt_create(Schema,Rules,Xtt),
	get_xtttable_name(Xmlobj,XttName),
	gateway_to_xtt1(Xmlobj,
			GatewayType,InEdges,OutEdges,XttName,Schema,Rules).

gateway_to_xtt1(_Xmlobj,GatewayType,InEdges,OutEdges,XttName,Schema,Rules):-
	gateway_type(exclusive,split,GatewayType),!,
	edges_atts(InEdges,In1),
	edges_conds_arg1_atts(OutEdges,In2),append(In1,In2,In),

	edges_atts(OutEdges,Outs),
	schema_create(XttName,In,Outs,Schema),

	rules_create(bpmn(GatewayType),
		     XttName,InEdges,OutEdges,Rules).


gateway_to_xtt1(_Xmlobj,GatewayType,InEdges,OutEdges,XttName,Schema,Rules):-
	gateway_type(exclusive,join,GatewayType),!,
	edges_atts(InEdges,Ins),
	edges_atts(OutEdges,Outs),
	schema_create(XttName,Ins,Outs,Schema),

	rules_create(bpmn(GatewayType),
		    XttName,InEdges,OutEdges,Rules).

gateway_to_xtt1(_Xmlobj,GatewayType,
	       InEdges,OutEdges,XttName,Schema,Rules):-
	gateway_type(parallel,split,GatewayType),!,
	edges_atts(InEdges,Ins),
	edges_atts(OutEdges,Outs),
	schema_create(XttName,Ins,Outs,Schema),

	rules_create(bpmn(GatewayType),XttName,InEdges,OutEdges,Rules).

gateway_to_xtt1(_Xmlobj,GatewayType,
	       InEdges,OutEdges,XttName,Schema,Rules):-
	gateway_type(parallel,join,GatewayType),!,
	edges_atts(InEdges,Ins),
	edges_atts(OutEdges,Outs),
	schema_create(XttName,Ins,Outs,Schema),

	%gateway_join_find_coupled_set(Xmlobj,CoupledSet),
	rules_create(bpmn(GatewayType),
		    XttName,InEdges,OutEdges,Rules).


%	Do I really need to implement it??? needed in inclusive join
%	gateways
% gateway_join_find_coupled_set(Xmlobj,CoupledEdges):-true.


xmlobj_create(Xml,Obj,XmlObject):-
	XmlObject = xmlobj(Xml,Obj).
xmlobj_xml(Xmlobj,Xml):-
	Xmlobj = xmlobj(Xml,_).
xmlobj_obj(Xmlobj,Obj):-
	Xmlobj = xmlobj(_,Obj).
xmlobj_name(Xmlobj,Name):-
	xmlobj_obj(Xmlobj,Obj),
	Obj = element(_,Atts,_),
	member('name'=Name,Atts),!.
xmlobj_name_nowhites(Xmlobj,Name):-
	xmlobj_name(Xmlobj,Name1),
	remove_white_non_spaces(Name1,Name),!.

xmlobj_edges(Xmlobj,Type,Edges):-
	Xmlobj = xmlobj(Xml,Node),
	node_edges_xml(Xml,Node,Type,XmlEdges),!,
	findall(Edg,
		(member(Edge,XmlEdges),xmlobj_create(Xml,Edge,Edgexmlobj),
		 xmlobj_to_attcond(Edgexmlobj,Att,Cond),
		 edge_create(Att,Cond,Edg)),
		Edges).

xmlobj_type(Xmlobj,Type):-
	xmlobj_obj(Xmlobj,Node),
	node_type(Node,Type).

xmlobj_to_att(Xmlobj,_Callback,Att):-
	xmlobj_obj(Xmlobj,Node),
	node_assigned_att(Node,Att),!.

xmlobj_to_att(Xmlobj,Callback,Att):-
	xmlobj_to_attname(Xmlobj,AttName),

	callback_attname(Callback,AttName),

	att_create(AttName,tpe_bpmn_entit,Callback,Att),

	xmlobj_obj(Xmlobj,Node),
	node_assign_att(Node,Att).

xmlobj_to_cond(Xmlobj,Cond):-
	xmlobj_obj(Xmlobj,Node),
	node_condexpr(Node,CondExpr),
	expr_create(CondExpr,Cond).

xmlobj_to_attcond(Xmlobj,Att,Cond):-
	callback_no(NoCallback),
	xmlobj_to_att(Xmlobj,NoCallback,Att),
	xmlobj_to_cond(Xmlobj,Cond).

xmlobj_to_attname(Xmlobj,AttName):-
	xmlobj_obj(Xmlobj,Node),
	node_type(Node,NodeType),
	node_type_to_attprefix(NodeType,AttPrefix),

	(\+ edge_type(NodeType),
	 get_xtt_name(xmlobj_name_nowhites(Xmlobj,AttName),
		     gensym(AttPrefix,AttName),
		     AttName)
	;
	 gensym(AttPrefix,AttName)
	),!.


:-dynamic node_attr/2.
node_clear:-retractall(node_attr(_,_)).

%%	node_condexpr(?Node,?CondExpr)	%%
%
%	Extracts conditional expression atom (CondExpr) from XML node (Node).
%
node_condexpr(Node,CondExpr):-
	node_type(Node,NodeType),edge_type(NodeType),!,
	node_attributes(Node,XmlAtts),
	(member('conditionType'='Default',XmlAtts),XmlCond=expr_cond_default,!
	;member('name'=XmlCond,XmlAtts),!
	;XmlCond=expr_no),
	CondExpr = XmlCond.

%%	node_id(+Node,-Id)	%%
%
%	Unique XML Node Id.
node_id(Node,Id):-
	Node = element(_,XmlAtts,_),
	member('xmi:id'=Id, XmlAtts),!,
	nonvar(Id).

node_assigned_att(Node,Att):-
	node_id(Node,Id),
	node_attr(Id,Att).

%%	node_assign_att(+Node,+Att)	%%
%
%	Causes side effect (calls assert).
%	Binds XML Node and its denoting XTT attribute (Att).
%
node_assign_att(Node,Att):-
	node_id(Node,Id),
	\+ node_attr(Id,Att),
	assert(node_attr(Id,Att)).

node_attributes(Node,Attributes):-
	Node = element(_,Attributes,_).

%% node_type_to_attprefix(?NodeType,?AttPrefix)	%%
%
%	XML node type (BPMN entity type)
%	and its corresponding attribute prefix.
%
node_type_to_attprefix(Type,a):-actv_type(Type),!.
node_type_to_attprefix(Type,e):-event_type(Type),!.
%node_type_to_attprefix(Type,g):-gateway_type(_,_,Type).
node_type_to_attprefix(Type,f):-edge_type(Type),!.


node_change_callbackname(Callback,ChangedCallback):-
    callback_name(Callback,CallbackName),
    callback_type(Callback,Type),callback_attname(Callback,AttName),
    callback_gen_name1(Type,CallbackName,AttName,CallbackName1),
    callback_create(Type,AttName,CallbackName1,ChangedCallback).

%%	node_edges_xml(+XML,+Node,+Type,-Edges)	%%
%
%	Extracts XML node (Node) Edges.
%
%	@param	Type	Edges type [in|out].
%	@param Edges	List of Edges terms (see edge_create/3).
%
node_edges_xml(XML,Node,Type,Edges):-
	node_edges_ids(Node,Type,Ids),
	findall(Edge,
		(member(Id,Ids),
		 node_find(XML,
			   'sequenceEdges',
			   ['xmi:id'=Id],
			   Edge)
		),Edges).

node_edges_ids(Node,Type,Ids):-
	Node = element(_,A,_),
	(
	 (Type = out,!, (member('outgoingEdges'=EdgesIds, A),! ; Ids=[]))
	;(Type = in ,!,	(member('incomingEdges'=EdgesIds, A),! ; Ids=[]))
	),
	parse_ids_of_edges(EdgesIds,Ids).

parse_ids_of_edges(EdgesIds, ParsedIds):-
       atomic_list_concat(ParsedIds,' ',EdgesIds).


edge_type(edge_type(_)).
actv_type(actv_type(_)).
event_type(event_type(_)).
gateway_type(GType,SplitOrJoin,gateway(GType,SplitOrJoin)).

%%	node_type(+Node,?Type)	%%
%
%	Obtains XML node (Node) type (Type).
%	@param	Type Denotes edge, activity, event or gateway type.
%
node_type(Node,EdgeType):-
	Node = element('sequenceEdges',_,_),!,
	edge_type(EdgeType).
node_type(Node,Type):-
	Node = element('vertices',XmlAtts,_),
	member('activityType'='Task',XmlAtts),!,
	actv_type(Type).
node_type(Node,Type):-
	node_gateway_type(Node,GatewayType),!,
	gateway_type(GatewayType,_,Type).
node_type(Node,Type):-
	Node = element('vertices',XmlAtts,_),
	member('activityType'=T,XmlAtts),
	atomic_list_concat(_,'Event',T),!,
	event_type(Type).

node_gateway_type(element('vertices',XmlAtts,_),exclusive):-
	member('activityType'='GatewayDataBasedExclusive',XmlAtts),!.
node_gateway_type(element('vertices',XmlAtts,_),parallel):-
	member('activityType'='GatewayParallel',XmlAtts),!.


gateway_sj_type(InEdges,OutEdges,split):-
	length(InEdges,1),!,
	length(OutEdges,N), N > 1.
gateway_sj_type(InEdges,OutEdges,join):-
	length(OutEdges,1),!,
	length(InEdges,N), N > 1.
gateway_sj_type(_,_,_):-
	write('gateway must be of split or join type'), fail.


node_actvs_all(Xml,NodesActvs):-
	node_find_all(Xml,'vertices',_,NodesActvs).


node_find(Xml,Name,Atts,Node):-
	Node = element(Name,Atts1,_),
	member(Node,Xml),
	member_all(Atts,Atts1).

node_find_all(Xml,Name,Atts,Nodes):-
	findall(Node,node_find(Xml,Name,Atts,Node),Nodes).

do_atts([],_,_,[]):-!.
do_atts([Att|Atts],Val,Handler,[Do|Dos]):-
	att_name(Att,AttName),
	copy_term(Handler,H1),
       	H1 = (AttName,Val,Do),
       	do_atts(Atts,Val,Handler,Dos).


%%	edge_create(?Att,?Cond,?Edge)	%%
%
%	Create BPMN Sequence Flow representing term (Edge).
%
%	@param	Att	Attribute term that denotes Edge (see att_create/4).
%	@param	Cond	Expression that denotes Sequence Flow condition (see expr_create/6).
%
edge_create(Att,Cond,Edge):-
	edge_att(Edge,Att),
	edge_cond(Edge,Cond).

edge_att(Edge,Att):-
	Edge = edge(Att,_).
edge_cond(Edge,Cond):-
	Edge = edge(_,Cond).

edge_cond_atts(Edge,Atts):-
	edge_cond(Edge,Cond),
	expr_atts(Cond,Atts).
edge_cond_arg1_att(Edge,Att):-
	edge_cond_atts(Edge,[Att|_]).


%%	edge_att_to_expristrue(?Edge,?IsTrueExpr)	%%
%
%	Transforms Edge to IsTrueExpr conditional
%	expression of Edge describing attribute
%	(see att_to_expristrue/2).
%
edge_att_to_expristrue(Edge,IsTrueExpr):-
	edge_att(Edge,Att),att_to_expristrue(Att,IsTrueExpr).

%%	edge_att_to_exprsettrue(?Edge,?AttSetExpr)	%%
%
%	Transforms Edge to AttSetExpr action expression
%	of Edge describing attribute
%	(see att_to_exprsettrue/2).
%
edge_att_to_exprsettrue(Edge,AttSetExpr):-
	edge_att(Edge,Att),att_to_exprsettrue(Att,AttSetExpr).

edge_to_exprs(InEdge,[Fin,InCond]):-
	edge_att_to_expristrue(InEdge,Fin),
	edge_cond(InEdge,InCond).

edges_atts(Edges,Atts):-
	list_extract(Edges,edge_att,Atts).
edges_conds(Edges,Cond):-
	list_extract(Edges,edge_cond,Cond).
edges_conds_atts(Edges,CondsAtts):-
	list_extract_flatten(Edges,edge_cond_atts,CondsAtts).
edges_conds_arg1_atts(Edges,CondArg1sAtts):-
	list_extract_flatten(Edges,edge_cond_arg1_att,CondArg1sAtts).

edges_atts_to_exprsettrue(Edges,AttsSetExprs):-
	findall(AttSetExpr,(
			    member(Edge,Edges),
			    edge_att_to_exprsettrue(Edge,AttSetExpr)),
	       AttsSetExprs).
edges_atts_to_expristrue(Edges,XttConds):-
	findall(XttCond,(
			    member(Edge,Edges),
			    edge_att_to_expristrue(Edge,XttCond)),
	       XttConds).

edges_to_exprs(InEdges,Exprs):-
	findall(Expr,(
				  member(InEdge,InEdges),
				  edge_to_exprs(InEdge,Expr)),
			Exprs1),
	flatten(Exprs1,Exprs).

edges_atts_and_conds_atts(Edges,Atts):-
	edges_atts(Edges,A1),edges_conds_atts(Edges,A2),
	append(A1,A2,Atts).


edges_default_as_last(Edges,EdgesLastDefault):-
	expr_create(expr_cond_default,CondDef),edge_create(_,CondDef,DefEdge),
	select(DefEdge,Edges,EdgesNoDef),!,
	append(EdgesNoDef,[DefEdge],EdgesLastDefault).
edges_default_as_last(Edges,Edges).


%%	xtt_create(?Schema,?Rules,?Xtt)	%%
%
%	XTT table representation.
%
%	@param	Schema	XTT table schema representation term (see schema_create/4).
%	@param	Rules	XTT table rules (see rule_create/6).
%
xtt_create(Schema,Rules,Xtt):-
	xtt_schema(Xtt,Schema),
	xtt_rules(Xtt,Rules).

xtt_schema(Xtt,Schema):-
	Xtt = xtt(Schema,_).
xtt_rules(Xtt,Rules):-
	Xtt = xtt(_,Rules).

xtts_schemas(Xtts,Schemas):-
	list_extract(Xtts,xtt_schema,Schemas).

xtts_ruless(Xtts,Ruless):-
	list_extract(Xtts,xtt_rules,Ruless).

xtts_rules_callbacks(Xtts,Callbacks):-
	xtts_ruless(Xtts,Ruless),
	findall(Callbacks1,(
			    member(Rules,Ruless),member(Rule,Rules),
			    rule_callbacks(Rule,Callbacks1)
			   ),
	       Callbackss),
	flatten(Callbackss,Callbacks1),
	list_to_set(Callbacks1,Callbacks).

xtts_write(Xtts,Stream):-
	xtts_schemas(Xtts,Schemas),
	forall(member(Schema,Schemas),(schema_write(Schema,Stream),stream_nl(Stream))),

	stream_nl(Stream),stream_nl(Stream),

	xtts_ruless(Xtts,Ruless),
	forall(member(Rules,Ruless),
	       forall( member(Rule,Rules),
		      (rule_write(Rule,Stream),stream_nl(Stream)))).

xtts_rules_callbacks_write(Xtts,Stream):-
	xtts_rules_callbacks(Xtts,Callbacks),
	callbacks_write(Callbacks,Stream).



%%%%%%%%	xtt schema	%%%%%%%%

%%	schema_create(?Name,?InAtts,?OutAtts,?Schema)	%%
%
%	Create Schema of XTT Table.
%
%	@param	Name	XTT table name.
%	@param	InAtts	Left hand side attributes.
%	@param	OutAtts	Right hand side attributes.
%
schema_create(Name,InAtts,OutAtts,Schema):-
	atts_names(InAtts,InNames),
	atts_names(OutAtts,OutNames),
	list_to_set(InNames,InNamesSet),
	list_to_set(OutNames,OutNamesSet),
	Schema = schema(Name,InNamesSet,OutNamesSet).

schema_write(Schema,Stream):-
	Schema = schema(Name,InNamesSet,OutNamesSet),

	%% xschm Name: InNamesSet ==> OutNamesSet.
	write(Stream,'xschm '),
	mywrite(Stream,Name),
	write(Stream,': '),
	mywrite(Stream,InNamesSet),
	write(Stream,' ==> '),
	mywrite(Stream, OutNamesSet),
	write(Stream,'.').


%%%%%%%%	xtt rule	%%%%%%%%

rule_create(XttName,Number,Conds,Acts,Rule):-
	rule_create(XttName,Number,Conds,Acts,[],Rule).

%%	rule_create(?XttName,?Number,+Conds,+Acts,+Callbacks,-Rule)
%
%	Creates Rule representation term.
%
%	@param	XttName	XTT table name.
%	@param	Number	Rule position number in XTT table.
%	@param	Conds	Rule conditions.
%	@param	Acts	Rule actions (executed if all conditions are satisfiable).
%	@param	Rule	Rule representing term.
%
rule_create(XttName,Number,Conds,Acts,Callbacks,Rule):-
	remove_empty(Conds,Conds1),
	rule_check_callbacks1(Callbacks),
	% ((Callbacks = [])
	 % -> callback_no(NoCallback), Callbacks1=[NoCallback]
	 % ; Callbacks1 = Callbacks),

	Rule = rule(XttName,Number,Conds1,Acts,Callbacks).

rule_xttname(Rule,XttName):-
	Rule = rule(XttName,_Number,_Conds,_Acts,_Callbacks).
rule_number(Rule,Number):-
	Rule = rule(_XttName,Number,_Conds,_Acts,_Callbacks).
rule_conds(Rule,Conds):-
	Rule = rule(_XttName,_Number,Conds,_Acts,_Callbacks).
rule_actions(Rule,Acts):-
	Rule = rule(_XttName,_Number,_Conds,Acts,_Callbacks).
rule_callbacks(Rule,Callbacks):-
	Rule = rule(_XttName,_Number,_Conds,_Acts,Callbacks1),

	((nonvar(Callbacks1))
		-> findall(C,(member(C,Callbacks1),\+callback_no(C)),Callbacks)
		; Callbacks1 = Callbacks).


rule_check_callbacks1(Callbacks):-var(Callbacks),!.
rule_check_callbacks1([]):-!.
rule_check_callbacks1([C|Cs]):-
	(callback_action(C);callback_no(C)),!,
	rule_check_callbacks1(Cs).
rule_check_callbacks1(_):-
	writeln('Rule callbacks is list of callbacks; each rule callback must be \'action\' or \'no\' type'), fail.

rule_any_callback(Rule):-
	rule_callbacks(Rule,Callbacks),
	member(C,Callbacks), \+ callback_no(C),!.

rule_write(Rule,Stream):-
	rule_xttname(Rule,XttName),
	rule_number(Rule,Number),
	rule_conds(Rule,Conds),
	rule_actions(Rule,Acts),
	rule_callbacks(Rule,Callbacks),

	% xrule XttName/Number: Conds1 ==> Acts.
	write(Stream,'xrule '),
	mywrite(Stream,XttName),write(Stream,'/'),mywrite(Stream,Number),
	write(Stream,': '),
	exprs_write(Conds,Stream), write(Stream,' ==> '), exprs_write(Acts,Stream),

	(\+ Callbacks = [],
	 % **> [[ on_start, [start] ]]
	 write(Stream,' **> '),
	 findall(CH,(member(C,Callbacks),
		     callback_name(C,CName),
		     callback_attname(C,CAttName),
		     CH=[CName,[CAttName]]
		    ),
		 CallbacksHeads),
	 mywrite(Stream,CallbacksHeads)
	;
	 true
	),!,

	write(Stream,'.').


%%%%	Gateways rules	%%%%%

rules_create(bpmn(GT),
	     XttName,InEdges,OutEdges,Rules):-
	gateway_type(exclusive,split,GT),!,
	rules_create_out_xor(XttName,InEdges,OutEdges,Rules).

rules_create(bpmn(GT),
	    XttName,InEdges,OutEdges,Rules):-
	gateway_type(exclusive,join,GT),!,
	rules_create_in_or(XttName,InEdges,OutEdges,Rules).

rules_create(bpmn(GT),
	    XttName,InEdges,OutEdges,Rules):-
	gateway_type(parallel,split,GT),!,
	rules_create_out_and(XttName,InEdges,OutEdges,Rules).

rules_create(bpmn(GT),
	    XttName,InEdges,OutEdges,Rules):-
	gateway_type(parallel,join,GT),!,
	rules_create_in_and(XttName,InEdges,OutEdges,Rules).



%%%%%%%%	in/out general rules	%%%%%%%%
%%
%%

%%	rules_create(+InOut,+LOp,?XttName,+InEdgesOrExpr,+OutEdgesOrExpr,-Rules)
%
%	Same as rules_create/7 with =|Opts = []|=.
%
rules_create(InOut,LOp,XttName,InEdgesOrExpr,OutEdgesOrExpr,Rules):-
	rules_create(InOut,LOp,XttName,InEdgesOrExpr,OutEdgesOrExpr,Rules,[]).


%%	rules_create(+InOut,+LOp,?XttName,+InEdgesOrExpr,+OutEdgesOrExpr,-Rules,+Opts)	%%
%
%	Creates BPMN entity input or output rules.
%	E.g. for generation rules that represents input edges logic 'and' connection
%	(all edges need to be active for further processing):
%	==
%	rules_create(in,and,
%		SomeXttName,InEdges,OutExpr,
%		Rules,[]).
%	==
%
%	@param	InOut			Type of rules [in|out].
%	@param	LOp				Edges logic connection operator see rules_logic_type/1.
%	Rules must represent this type of logic connection. Rules must belong to same XTT table;
%	i.e. for 'and' type rules have to represent 'and' condition
%	for 'or' type rules have to represent 'or' condition etc.
%
%	@param	XttName			XTT table name.
%	@param	InEdgesOrExpr	Input edges or input expression.
%	@param	OutEdgesOrExpr	Output edges or output expression.
%	@param	Rules			Created rules.
%	@param	Options			List of options; so far only one is supported: nocallbacks -- forces to not generate rules callbacks (actions).
%
%
rules_create(InOut,LOp,XttName,InEdgesOrExpr,OutEdgesOrExpr,Rules,Options):-
		rules_create1(InOut,LOp,XttName,InEdgesOrExpr,OutEdgesOrExpr,Rules),

		((member(nocallbacks,Options))
			-> rules_callbacks_one(Rules,[])
			; true).

rules_create1(in,and,XttName,InEdges,OutExprs,[Rule]):-
	edges_to_exprs(InEdges,InExprs),
	rule_create(XttName,_,InExprs,OutExprs,_Callbacks,Rule).

rules_create1(in,or,_,[],_,[]):-!.
rules_create1(in,or,XttName,
		    [InEdge|InEdges],OutExprs,[Rule|Rules]):-
	edge_to_exprs(InEdge,InExprs),
	rule_create(XttName,_N,InExprs,OutExprs,_Callbacks,Rule),

	rules_create1(in,or,XttName,InEdges,OutExprs,Rules).



rules_create1(out,and,XttName,InConds,OutEdges,[Rule]):-
	edges_atts_to_exprsettrue(OutEdges,FoutExprs),
	rule_create(XttName,_N,InConds,FoutExprs,_Callbacks,Rule).

rules_create1(out,or,_,_,[],[]):-!.
rules_create1(out,or,XttName,
		     InExprs,[OutEdge|OutEdges],[Rule|Rules]):-
	edge_att_to_exprsettrue(OutEdge,FoutExpr),
	rule_create(XttName,_N,InExprs,[FoutExpr],_Callbacks,Rule),

	rules_create1(out,or,XttName,InExprs,OutEdges,Rules).

rules_create1(out,xor,XttName,InExprs,OutEdges,Rules):-
	edges_default_as_last(OutEdges,OutEdges1),
	rules_create_gexs(XttName,1,InExprs,OutEdges1,Rules).


%	rules_create_gexs(+XttName,+N,+PreviousConds,+OutEdgesDefaultLast,-Rules).
%
%	Special rules creation for xor split.
%	Default edge (that have default condition) must be last element
%	of OutEdges simply call
%	edges_default_as_last/2 before
%	calling this predicate.
%
rules_create_gexs(_,_,_,[],[]):-!.
rules_create_gexs(XttName,N,
		  PreviousConds,[OutEdge|OutEdgesDefaultLast],[Rule|Rules]):-
	edge_cond(OutEdge,CurrentCond),

	(expr_type(CurrentCond,expr_cond_default),Conds = PreviousConds,!
	; append(PreviousConds,[CurrentCond],Conds)),

	edge_att_to_exprsettrue(OutEdge,FoutExpr),
	rule_create(XttName,N,Conds,[FoutExpr],Rule),

	N1 is N + 1,
	expr_arg1_name(FoutExpr,FoutAttName),
	expr_create(expr_cond,FoutAttName,null,eq,NotFout),
	append(PreviousConds,[NotFout],PreviousConds1),
	rules_create_gexs(XttName,N1,PreviousConds1,OutEdgesDefaultLast,Rules).


%%%%%%%%	Short cuts rules creation predicates	%%%%%%%%
%	used by gateways rules creation predicates
%	see above.
%
rules_create_in_and(XttName,InEdges,[OutEdge],Rules):-
	edge_att_to_exprsettrue(OutEdge,Fout),
	rules_create(in,and,XttName,InEdges,[Fout],Rules,[nocallbacks]),
	rules_ensure_numbers(Rules,1).

rules_create_in_or(XttName,InEdges,[OutEdge],Rules):-
	edge_att_to_exprsettrue(OutEdge,FoutExpr),
	rules_create(in,or,XttName,InEdges,[FoutExpr],Rules,[nocallbacks]),
	rules_ensure_numbers(Rules,1).

rules_create_out_and(XttName,[InEdge],OutEdges,Rules):-
	edge_att_to_expristrue(InEdge,FinCond),
	rules_create(out,and,XttName,[FinCond],OutEdges,Rules,[nocallbacks]),
	rules_ensure_numbers(Rules,1).

rules_create_out_xor(XttName,[InEdge],OutEdges,Rules):-
	edge_att_to_expristrue(InEdge,FinCond),
	rules_create(out,xor,XttName,[FinCond],OutEdges,Rules,[nocallbacks]),
	rules_ensure_numbers(Rules,1).


%%	rules_create_in_out(+InLOp,+OutLOp,+XttName,+InEdges,+OutEdges,+Att,+Clbks,-Rules)
%
%	Creates input and output rules for BPMN entity that is denoted by Att attribute.
%
%	@param	InLOp	Input edges logic operator type see rules_create/6.
%	@param	OutLOp	Output edges logic operator type see rules_create/6.
%	@param	XttName	XTT table name to which Rules belong.
%	@param	Att	Attribute that denotes BPMN entity.
%	@param	InEdges	BPMN entity input edges.
%	@param	OutEdges	BPMN entity output edges.
%	@param	Att	BPMN entity denoting attribute.
%	@param	Clbks	Callback that is called when BPMN entity becomes active.
%	@param	Rules	Created rules.
%
rules_create_in_out(InLOp,OutLOp,XttNm,InEdges,OutEdges,Att,Clbks,Rules):-
	rules_logic_type(InLOp),rules_logic_type(OutLOp),

	att_to_exprsettrue(Att,OutExpr),
	% expr_cond_equiv_action(InExpr,OutExpr,Att),

	rules_create(in,InLOp,XttNm,InEdges,[OutExpr],InRules),
	rules_callbacks_one(InRules,Clbks),

	(InRules = [], InExprs = []
	; InExprs = [InExpr], expr_cond_equiv_action(InExpr,OutExpr,Att)),!,

	MiddleRules = [MidRule],
	rule_create(XttNm,_,InExprs,[],Clbks,MidRule),

	rules_create(out,OutLOp,XttNm,InExprs,OutEdges,OutRules,[nocallbacks]),


	rules_in_out_optimize(InRules,MiddleRules,OutRules,Rules),
	rules_ensure_numbers(Rules,1).


%%	rules_logic_type(?Type)	%%
%
%	Type is allowed type for any rules set.
%	Rules set have to represent some edges logic connection
%	see rules_create/6.
%
%	@param	Type	[and | or | xor]
%
rules_logic_type(and).
rules_logic_type(or).
rules_logic_type(xor).


%%%% Below there is a tricky code that optimizes (merges)
%%%% in and out rules where it is possible

rules_in_out_optimize([InRule],[MidRule],OutRules,Rules):-!,
	rules_in_out_optimize1([InRule],[MidRule],OutRules,Rules).

rules_in_out_optimize(InRules,[MidRule],[OutRule],Rules):-!,
	rules_in_out_optimize([MidRule],[OutRule],Rules1),
	rules_in_out_optimize(InRules,Rules1,Rules).

rules_in_out_optimize(InRules,MidRules,OutRules,Rules):-
	rules_in_out_optimize1(InRules,MidRules,OutRules,Rules).

rules_in_out_optimize1(InRules,MidRules,OutRules,Rules):-
	rules_in_out_optimize(InRules,MidRules,Rules1),
	rules_in_out_optimize(Rules1,OutRules,Rules).


rules_in_out_optimize([],[],[]):-!.
rules_in_out_optimize(InRules,[],InRules):-!.
rules_in_out_optimize([],OutRules,OutRules):-!.
rules_in_out_optimize(InRules,[OutRule|OutRules],Rules):-
	rules_in_out_optimize0(InRules,OutRule,InRulesOpt),
	rules_in_out_optimize(InRulesOpt,OutRules,Rules).

rules_in_out_optimize0(InRules,OutRules,Rules):-
	(rules_in_out_optimize01(InRules,OutRules,Rules)
	;rules_in_out_optimize02(InRules,OutRules,Rules)),!.

rules_in_out_optimize01([],_,[]):-!.
rules_in_out_optimize01([InRule|InRules],OutRule,[Rule|Rules]):-
	rules_in_out_optimize1([InRule],[OutRule],[Rule]),!,
	rules_in_out_optimize01(InRules,OutRule,Rules).

rules_in_out_optimize02(InRules,OutRule,Rules):-
	append(InRules,[OutRule],Rules).


%	rules_in_out_optimize1(+InRule,+OutRule,-Rules)
%
%	Optimize core predicate.
%
rules_in_out_optimize1([InRule],[OutRule],[Rule]):-
	rule_conds(InRule,InConds),rule_actions(InRule,InActions),rule_callbacks(InRule,InClbks),
	rule_conds(OutRule,OutConds),rule_actions(OutRule,OutActions),rule_callbacks(OutRule,OutClbks),

	(
	 forall(member(OutCond,OutConds),(member(InAction,InActions),
					  expr_cond_equiv_action(OutCond,InAction,_Att)))
	;
	 forall(member(OutCond,OutConds),(member(InCond,InConds),
					  InCond = OutCond))
	),!,

	rule_xttname(InRule,XttName),

	append(InClbks,OutClbks,Clbks),list_to_set(Clbks,ClbksSet),
	append(InActions,OutActions,Actions),list_to_set(Actions,ActionsSet),
	rule_create(XttName,_,InConds,ActionsSet,ClbksSet,Rule),!.

rules_in_out_optimize1([In],[Out],[In,Out]).




rules_callbacks_one([],_):-!.
rules_callbacks_one([Rule|Rules],Callbacks):-
	rule_callbacks(Rule,Callbacks),
	rules_callbacks_one(Rules,Callbacks).

rules_ensure_numbers(Rs,N):-
	rules_ensure_numbers(Rs,N,_).

rules_ensure_numbers([],NLN,NLN):-!.
rules_ensure_numbers([R|Rs],N,NLN):-
	rule_number(R,N),
	N1 is N + 1,
	rules_ensure_numbers(Rs,N1,NLN).




%%%%	Callbacks	%%%%
%
%	Type = {demand|action|no}
%	demand	- prefix:get
%	action	- prefix:on
%	no		- no callback

%%	callback_create(?Type,?AttName,?Name,?Callback)	%%
%
%	Creates callback representing term.
%
%	@param	Type		Callback type:
%						$ no:
%							Callback is term that means 'no callback'.
%						$ demand:
%							Attribute callback; attribute value supplier (Heart callback).
%						$ action:
%							Rule callback; on successfully executed rule (Heart action).
%
%	@param	AttName		Attribute name that callback corresponds to.
%	@param	Name		Callback name.
%	@param	Callback	Term that represents callback.
%
callback_create(Type,AttName,Name,Callback):-
	callback_type(Callback,Type),
	callback_attname(Callback,AttName),
	callback_name(Callback,Name).

callback_create_genname(Type,AttName,NameCore,Callback):-
	callback_gen_name1(Type,NameCore,Name),
	callback_create(Type,AttName,Name,Callback).

callback_no(callback(no,_,_)).
callback_action(callback(action,_,_)).
callback_demand(callback(demand,_,_)).

callback_type(Callback,Type):-
	Callback = callback(Type,_AttName,_Name).
callback_attname(Callback,AttName):-
	Callback = callback(_Type,AttName,_Name).
callback_name(Callback,Name):-
	Callback = callback(_Type,_AttName,Name).

callback_type_prefix1(action,on).
callback_type_prefix1(demand,get).

callback_gen_name1(_CallbackType,OriginalName,_AttName,CallbackName):-
	var(OriginalName), CallbackName = OriginalName,!.
callback_gen_name1(CallbackType,OriginalName,AttName,CallbackName):-
	get_xtt_name(OriginalName,
                 callback_gen_name1(CallbackType,AttName,CallbackName),
                 CallbackName).

callback_gen_name1(Type,NameCore,CallbackName):-
	callback_type_prefix1(Type,Prefix),
	callback_gen_name2(Prefix,NameCore,CallbackName).

callback_gen_name2(Prefix,NameCore,CallbackName):-
	atomic_list_concat([Prefix,NameCore],'_',CallbackName).

callback_to_xttcallback(Callback,XttCallback):-
	callback_name(Callback,CallbackName),
	callback_attname(Callback,AttName),
	callback_to_xttcallback1(CallbackName,AttName,XttCallback).
callback_to_xttcallback1(CallbackName,AttName,XttCallback):-
	atomic(CallbackName),atomic(AttName),!,
	XttCallback = [CallbackName,[AttName]].
callback_to_xttcallback1(_,_,[]).


callback_write(Callback,Stream):-
	\+ callback_no(Callback),!,

	callback_type(Callback,Type),
	callback_name(Callback,CallbackName),
	callback_attname(Callback,AttName),


	% [xcall|xactn] CallbackName : [AttName] >>> Goal.
	(Type = demand, write(Stream,'xcall ')
	;
	 Type = action, write(Stream,'xactn ')
	),!,

	mywrite(Stream,CallbackName),write(Stream,': '),mywrite(Stream,[AttName]),
	write(Stream,' >>> '),

	(Type = demand,
	 % att_name(Att,AttName),node_attr(_,Att),
	 callback_write1(Stream,CallbackName,AttName)
	;
	 Type = action,
	 write(Stream, 'writeln('), mywrite(Stream,CallbackName),
	 write(Stream,'). %%% Put your code here; use alsv_new_val(AttName,Answer) to set attribute value.' )
	),
	!,
	stream_nl(Stream).

callback_write(_,_).

callback_write1(Stream,CallbackName,AttName):-
	  write(Stream,'(writeln('),mywrite(Stream,CallbackName),write(Stream,'),'),
	  write(Stream,'read(Answer), alsv_new_val('),
	  mywrite(Stream,AttName),
	  write(Stream,',Answer)).').

callbacks_write(Callbacks,Stream):-
	forall(member(Callback,Callbacks),(callback_write(Callback,Stream))).


%%%%	Types of Attributes	%%%%
%
%


:-dynamic attrib_type/1.

typeatt_create(Name,Length,Scale,Domain,TypeAtt):-
	TypeAtt = typeatt(Name,symbolic,Length,Scale,'',Domain).

typeatt_name(TypeAtt,Name):-
	TypeAtt = typeatt(Name,_Base,_Length,_Scale,_Description,_Domain).

typeatt_write(TypeAtt,Stream):-
	TypeAtt = typeatt(Name,Base,Length,Scale,Description,Domain),

	write(Stream,xtype),write(Stream,' '),
	writeq(Stream,[
		name:Name,
		base:Base,
		length:Length,
		scale:Scale,
		desc:Description,
		domain:Domain]),
	write(Stream,'.').

typeatt_add(TypeAtt):-
	\+ attrib_type(TypeAtt),
	assert(attrib_type(TypeAtt)).
typeatt_replace(TypeAtt):-
	att_name(TypeAtt,Name),
	att_name(TA,Name),
	retractall(attrib_type(TA)),
	assert(attrib_type(TypeAtt)).
typeatt_get(TypeAtt):-
	attrib_type(TypeAtt).

typeatt_add_domain_values(TypeName,Values):-
	TypeAtt = typeatt(TypeName,Base,Length,Scale,Domain),
	typeatt_get(TypeAtt),
	append(Domain,Values,D1),
	list_to_set(D1,NewDomain),
	typeatt_replace(typeatt(TypeName,Base,Length,Scale,NewDomain)).


typeatt_name_att_name(integer,integer):-!.
typeatt_name_att_name(TypeAttName,AttName):-
	atomic_list_concat([tpe_,AttName],TypeAttName).

% Some predefined types
:-typeatt_add(typeatt(tpe_bpmn_entit,
		      symbolic,
		      4,
		      0,
		      'bpmn entity 1 - executed, null - not executed',
		      [null,1])
	     ).

:-op(300,xfy,to).
:-typeatt_add(typeatt(integer,numeric,10,0,integer,[-2147483648 to 2147483647])).


%%%%	Attributes	%%%%
%
%
:-dynamic attrib/1.
att_clear:-retractall(attrib(_)).

%%	att_register(+Att)	%%
%
%	Registers Att based on Name and Callback parameters.
att_register1(Att):-
	att_name(Att,Name),att_name(Att2,Name),
	((\+ attrib(Att2))
		-> assert(attrib(Att))
		;  Att = Att2).
att_register1(Att):-
	swritef(S,'Attribute: ~w can not be registered;
attribute with same name already exists
and has different properties.', [Att]),

	string_to_atom(S,M),
	throw(error(att_regiter/1,M)).

%%	att_exists(?Att)	%%
%
%	Check if Att is registered.
att_exists(Att):-
	attrib(Att),!.

%%	att_exists(?AttName,?Att)	%%
%
%	Same as att_exists/1 =|att_name(Att,AttName)|=.
att_exists(AttName,Att):-
	att_name(Att,AttName),
	att_exists(Att).


%%	att_create(?Name,?TypeName,?Callback,?Att)	%%
%
%	Creates and registers attribute representing term.
%
%	@param	Name		Attribute name.
%	@param	TypeName	Attribute type name.
%	@param	Callback	Attribute callback term.
%	@param	Att			Attribute term.
%
att_create(Name,TypeName,Callback,Att):-
	att_name(Att,Name),
	att_typename(Att,TypeName),
	att_callback(Att,Callback),
	( att_exists(Att)
	; att_register1(Att)),!.

att_create(Name,_Type,_CallbackName,_Att):-
	swritef(M,'Can not create attribute: ~w', [Name]),
	throw(error(att_create/4,M)).

%%	att_name(?Att,?Name)	%%
%
%	Extracts attribute name.
%
%	@param Att	attribute.
%	@param Name	name of attribute Att.
%
att_name(Att,Name):-
	Att=att(Name,_,_,_,_,_).

%%	att_abbrev(+Att,-Abbrev)	%%
%
%	@param Abbrev	read only (generated from att_name(Att,AttName).
%
att_abbrev(Att,Abbrev):-
	% Att=att(_,Abbrev1,_,_,_,_),
	att_name(Att,Name),
	atomic_list_concat([Name,1],Abbrev).

att_class(Att,Class):-
	Att=att(_,_,Class,_,_,_).
att_typename(Att,TypeName):-
	Att=att(_,_,_,TypeName,_,_).
att_comm(Att,Comm):-
	Att=att(_,_,_,_,Comm,_).
att_callback(Att,Callback):-
	Att=att(_,_,_,_,_,Callback).

att(Att):-
	att_name(Att,_).


att_get(AttName,Att):-
	Att = att(AttName,_,_,_,_,_),
	attrib(Att).

%%	att_to_expriseq(?Att,?Val,?EqExpr)	%%
%
%	Transforms attribute to is equal expression.
%
%	@param	Att		Attribute.
%	@param	Val		Value.
%	@param	EqExpr	Expression term which means 'Att equals Val'.
%
att_to_expriseq(Att,Val,EqExpr):-
	att_name(Att,BAttName),
	% BAttName eq Val
	expr_create(expr_cond,BAttName,Val,eq,
				EqExpr).

%%	att_to_expristrue(?Att,?IsTrueExpr)
%
%	Creates expression that means 'Att equals true'.
%
att_to_expristrue(Att,IsTrueExpr):-
	att_to_expriseq(Att,1,IsTrueExpr).

%%	att_to_exprattset(?Att,?Val,?ExprAttSet)
%
%	Creates expression that means 'assign Val to Att'.
%
att_to_exprattset(Att,Val,ExprAttSet):-
	att_name(Att,AttName),
	% AttName set Val
	expr_create(expr_actn,AttName,Val,set,
				ExprAttSet).

%%	att_to_exprsettrue(?Att,?AttSetTrue)
%
%	Creates expression that means 'assign true to Att'.
%
att_to_exprsettrue(Att,AttSetTrue):-
	att_to_exprattset(Att,1,AttSetTrue).

att_to_xttatt(Att,XttAtt):-
	XttAtt1 = [name:Name,abbrev:Abbrev,
		   class:Class,type:TypeName,comm:Comm],

	att_name(Att,Name),att_abbrev(Att,Abbrev),att_class(Att,Class),
	att_typename(Att,TypeName),att_comm(Att,Comm),

	att_callback(Att,Callback),
	((\+ callback_no(Callback))
          -> (callback_to_xttcallback(Callback,XttCallback),
	      append(XttAtt1,[callback: XttCallback],XttAtt))
          ; XttAtt = XttAtt1).

att_write(Att,Stream):-
	write(Stream,'xattr['),

	att_to_xttatt(Att,XttAtt),
	att_write_helper(XttAtt,Stream),
	
	write(Stream, '].'),stream_nl(Stream).

att_write_pair(Type:Value,Stream):-
	att_write_pair_last(Type:Value,Stream), write(Stream,,).
att_write_pair_last(Type:Value,Stream):-
	mywrite(Stream, Type), mywrite(Stream,:), mywrite(Stream,Value).
	
att_write_helper([Pair],Stream):-
	att_write_pair_last(Pair,Stream).
att_write_helper([Pair|T],Stream):-
	att_write_pair(Pair,Stream),
	att_write_helper(T,Stream).

atts_names([],[]):-!.
atts_names(Atts,Names):-
	findall(Name, (member(Att,Atts),att_name(Att,Name)), Names).


atts_finalize(Atts,AttsFinalized):-
	copy_term(Atts,AttsFinalized),
	atts_finalize1(AttsFinalized).

atts_finalize1([]):-!.
atts_finalize1([Att|Atts]):-
	att_class(Att,simple),

	att_typename(Att,AttName),(
	var(AttName),AttName=integer,!;true),

	att_comm(Att,inter),
	atts_finalize1(Atts).


atts_callbacks(Atts,Callbacks):-
	findall(Callback,
	        (member(Att,Atts),att_callback(Att,Callback)),
		Callbacks).


atts_write(Atts,Stream):-
	forall(member(Att,Atts),att_write(Att,Stream)).

atts_callbacks_write(Atts,Stream):-
	atts_callbacks(Atts,Callbacks),
	callbacks_write(Callbacks,Stream).





%%%%	Expressions	%%%%
%
%	Type = {expr_no|expr_cond_default|expr_cond|expr_action}

expr_check_type1(Type):-
	( Type=expr_no	; Type=expr_cond_default
	; Type=expr_cond; Type=expr_actn),!.


%%	expr_create(?Type,+Arg1Name,+Arg2Name,?Operator,?Expr)	%%
%
%	Same as expr_create/6 with =|Atts=_|=.
expr_create(Type,Arg1Name,Arg2Name,Operator,Expr):-
	expr_create(Type,Arg1Name,Arg2Name,Operator,_Atts,Expr).

%%	expr_create(?Type,+Arg1Name,+Arg2Name,?Operator,-Atts,?Expr)
%
%	expr_create(?Type,-Arg1Name,-Arg2Name,?Operator,+Atts,?Expr)
%
%	Creates expression (Expr) that is one of following types (Type):
%
%	|	*Values*			|	*Meaning*													|
%	|	expr_no				|	no expression												|
%	|	expr_cond_default		|	used to denote default sequence flow (default condition)	|
%	|	expr_cond			|	conditional expression (condition)							|
%	|	expr_action			|	action e.g. assigning value to attribute					|
%
%	@param	Type	Expression type see above table.
%	@param	Arg1Name	Expression left side argument.
%	@param	Arg2Name	Expression right side argument.
%	@param	Operator	Expression operator.
%	@param	Atts		List of expression attributes, corresponding to Arg1Name and Arg2Name.
%	If Arg2Name denotes attribute =|(Arg2Name = ${Att2Name})|= then =|Atts = [Att1,Att2]|=
%	else =|Atts = [Att1]|=.
%
%
%
expr_create(Type,Arg1Name,Arg2Name,Operator,Atts,Expr):-
	expr_type(Expr1,Type),
	expr_arg1_name(Expr1,Arg1Name),
	expr_arg2_name(Expr1,Arg2Name),
	expr_op(Expr1,Operator),
	expr_ensure_atts(Expr1,Expr,Atts).


%%	expr_create(+XmlExpr,-Expr)	%%
%
%	Create term that denotes expression. Four cases exists.
%
%	@param	XmlExpr		XmlExpr is an atom that is in one of following forms:
%						$ Arg1 Op Arg2 :
%							Arg1 is expression attribute name  =|att_name(Att1,Arg1)|= (see att_name/2).
% Op is operator name see op_to_op/2. If =|Arg2 = ${Att2Name}|= then Att2Name denotes attribute (Att2) current value
% else Arg2 denotes explicit value of Att1.
%						$ Arg1 :
%							Expr is existential expression see expr_existential/1.
%						$ expr_no :
%							Expr is created by expr_no/1.
%						$ expr_cond_default :
%							Expr is created by expr_cond_default/1.
%
%	@param	Expr		Term that represents expression see expr_create/6 for details.
%
expr_create(XmlExpr,Expr):-
	expr_parse_xml(XmlExpr,Arg1Name,Arg2Name,Operator),!,

	expr_op_type1(Operator,Type),
	expr_create(Type,Arg1Name,Arg2Name,Operator,_,Expr).


expr_create(expr_no,NoExpr):-!,
	expr_no(NoExpr).

expr_create(expr_cond_default,ExprDefault):-!,
	expr_cond_default(ExprDefault).

expr_create(XmlExpr,Expr):-
	atomic(XmlExpr),!,

	expr_arg1_name(Expr1,XmlExpr),
	expr_existential(Expr1),
	expr_ensure_atts(Expr1,Expr,_Atts).

%%	expr_no(?NoExpr)	%%
%
%	Creates term that denotes lack of expression.
%
%	@param	NoExpr	Term that means 'no expression'.
%
expr_no(expr(expr_no,nothing_,nothing_,nothing_,[])).

%%	expr_cond_default(?ExprDefault)	%%
%
%	Creates term that denotes default sequence flow (default condition) (see BPMN reference for details).
%
%	@param	ExprDefault	Default condition expression term.
%
expr_cond_default(expr(expr_cond_default,nothing_,nothing_,nothing_,[])).

%%	expr_type(?Expr,?Type)	%%
%
%	@param Expr	Expression term.
%	@param	Type	Expression type term (see expr_create/6).
%
expr_type(Expr,Type):-
	Expr = expr(Type,_AttName1,_AttName2,_Operator,_Atts),
	expr_check_type1(Type).
expr_arg1_name(Expr,AttName1):-
	Expr = expr(_Type,AttName1,_AttName2,_Operator,_Atts).
expr_arg2_name(Expr,AttName2):-
	Expr = expr(_Type,_AttName1,AttName2,_Operator,_Atts).
expr_op(Expr,Operator):-
	Expr = expr(_Type,_AttName1,_AttName2,Operator,_Atts).
expr_atts(Expr,Atts):-
	Expr = expr(_Type,_AttName1,_AttName2,_Operator,Atts).
expr_elems(Expr,Type,Arg1,Arg2,Op,Atts):-
	Expr = expr(Type,Arg1,Arg2,Op,Atts).

expr_arg_name(Expr,ArgNum,ArgName):-
	( ArgNum = 1, expr_arg1_name(Expr,ArgName)
	; ArgNum = 2, expr_arg2_name(Expr,ArgName)),!.

expr_arg2_expr(Expr,Arg2Expr):-
	expr_arg2_name(Expr,Arg2Name),
	(expr_atts(Expr,[_,_]), Arg2Expr = ${Arg2Name}
	;
	 Arg2Expr = Arg2Name),!.

expr_att1(Expr,Att):-
	expr_arg1_name(Expr,AttName),
	att_get(AttName,Att).
expr_att2(Expr,Att):-
	expr_arg2_name(Expr,AttName),
	att_get(AttName,Att).

expr_cond_equiv_action(Cond,Action,Att):-
	att_to_exprsettrue(Att,Action),att_to_expristrue(Att,Cond).

expr_to_xttexpr(Expr,[]):-
	Expr = expr(Type,_,_,_),
	(Type = expr_no ; Type = expr_cond_default),!.
expr_to_xttexpr(Expr,ALSVExpr):-
	Expr = expr(_Type,A,B,Op),
	%ALSVExpr = (A Op A).
	atomic_list_concat([A,Op,B],' ', ALSVExpr1),
	term_to_atom(ALSVExpr,ALSVExpr1).

expr_parse_xml(XmlExpr,Arg1,Arg2,Operator):-
	%atomic_list_concat([Arg1,Operator1,Arg2],' ',XmlExpr),
	atom_codes(XmlExpr,ExprS),
	phrase(arg_op_arg(Arg11,Operator11,Arg21),ExprS,_R),!,
	\+ Operator11=[], \+ Arg21 = [],
	read_from_chars(Arg11,Arg1),
	read_from_chars(Operator11,Operator1),
	read_from_chars(Arg21,Arg2),
	op_to_op(Operator1,Operator).


% expr_op_type1(Operator,Type)
expr_op_type1(set,expr_actn):-!.
expr_op_type1(_,expr_cond):-!.

arg_op_arg(Arg1,Op,Arg2) -->
	whites(_W0),
	non_whites(Arg1),
	whites(_W1),
	non_whites(Op),
	whites(_W2),
	non_whites(Arg2).

%% op_to_op(+Op,-ALSVOp)
%
%	op_to_op(-Op,+ALSVOp)
%
%	Translates Op to ALSVOp:
%
%	| *Op*	| *ALSVOp* |
%	|	=	|	eq	|
%	|	< 	|	lt	|
%	|	> 	|	gt	|
%	|	<=	|	lte	|
%	|	>=	|	gte	|
%
%	For all others Op values that are not listed above =| Op = ALSVOp |= which means that
%	if =|Op = gt|= then =|ALSVOp = gt|= also.
%
op_to_op('=','eq'):-!.
op_to_op('<','lt'):-!.
op_to_op('>','gt'):-!.
op_to_op('<=','lte'):-!.
op_to_op('>=','gte'):-!.
op_to_op(Op,Op).  % in case Op is already XttOp then translation is not needed

%% op_neg(?Op,?NegatedOp)
%
%	@param Op		Operator.
%	@param NegatedOp	Negated operator e.g. if =|Op = eq|= then =|NegatedOp = neq|=.
%
op_neg(Op,NegatedOp):-
	(op_neg1(Op,NegatedOp);op_neg1(NegatedOp,Op)),!.

op_neg(Op,not(Op)). % default one just to denote negation

op_neg1(eq,neq).
op_neg1(in,notin).
op_neg1(lt,gte).
op_neg1(gt,lte).

expr_neg(Expr,Expr):-
	expr_type(Expr,Type),
	(Type = expr_no ; Type = expr_cond_default),!.

expr_neg(Expr,NegExpr):-
	expr_create(expr_cond,Arg1,Arg2,Op,Expr),
	op_neg(Op,NegOp),
	expr_create(expr_cond,Arg1,Arg2,NegOp,NegExpr).

%%	expr_existential(?Expr)
%
%	Expr matches existential expression which means
%	'Arg1 is any' or 'does Arg1 exists in knowledge base',
%	where =|expr_arg1_name(Expr,Arg1)|= .
%
%	@param Expr	Existential expression.
%
expr_existential(Expr):-
	expr_create(expr_cond,_,any,eq,Expr).

expr_ensure_atts(Expr,Expr,[]):-
	expr_no(Expr),!;
	expr_cond_default(Expr).

expr_ensure_atts(Expr,OExpr,Atts):-
	expr_elems(Expr,Type,Arg1,Arg2,Op,_Atts1),
	nonvar(Arg1),nonvar(Arg2),!,

	expr_elems(OExpr,Type,Arg1,Arg2N,Op,Atts),
	expr_ensure_att(Expr,1,Att1),

	% if Arg2 = ${AttName} then Arg2 is attribute name
	% else Arg2 is just value of Arg1 which is in Att1 domain
	( Arg2 = ${Arg2N},
	  expr_ensure_att(OExpr,2,Att2),
	  list_to_set([Att1,Att2],Atts)
	 ;
	  Arg2N = Arg2, Atts = [Att1]
	),!,
	expr_atts(OExpr,Atts).

expr_ensure_atts(Expr,Expr,Atts):-
	list_to_set(Atts,Atts1),
	expr_atts(Expr,Atts1),
	nonvar(Atts1),
	[Att1|R] = Atts1,

	expr_ensure_att(Expr,1,Att1),
	( R=[Att2], expr_ensure_att(Expr,2,Att2)
	 ; true),!.



expr_ensure_att(Expr,ArgNum,Att):-
	expr_arg_name(Expr,ArgNum,AttName),
	att_exists(AttName,Att),!.

expr_ensure_att(Expr,ArgNum,Att):-
	expr_arg_name(Expr,ArgNum,AttName),

	callback_create_genname(demand,AttName,AttName,Callback1),
	((is_callback_only_for_existential_exprs)
		-> ((expr_existential(Expr)) -> Callback = Callback1 ; callback_no(Callback))
		;  Callback = Callback1),


	att_create(AttName,_AttTypeName,Callback,Att).


expr_to_xttstring(Expr,String):-
	expr_type(Expr,Type),
	\+ Type = expr_no, \+ Type = expr_cond_default,

	expr_arg1_name(Expr,Arg1Name),
	expr_arg2_expr(Expr,Arg2Expr),
	expr_op(Expr,Op),

	%  Arg1Name' 'Op' 'Arg2Expr

	atom_to_codes_qout_tailspace(Arg1Name,C1,C2),
	atom_to_codes_qout_tailspace(Op,C2,C3),
	atom_to_codes_qout(Arg2Expr,C3,[]),

	string_to_list(String,C1).


exprs_write(Exprs,Stream):-
	findall(XttExpr,(member(Expr,Exprs),
			 expr_to_xttstring(Expr,XttExpr)),
	       XttExprs),
	write(Stream,XttExprs).

exprs_atts1names(Exprs,Atts1Names):-
	findall(Att1Name,(member(Expr,Exprs),expr_arg1_name(Expr,Att1Name)),
	       Atts1Names).



get_xtttable_name(Xmlobj,Name):-
	is_bpmn_atts_names,
	(xmlobj_name_nowhites(Xmlobj,Name1),
	 atom_codes(Name1,Codes),length(Codes,N),N>0
	;
	 xmlobj_type(Xmlobj,Type),
	 Type =..L, nth0(0,L,Base),
	 gensym(Base,Name1)
	),!,
	atomic_list_concat(['xtt',Name1],'_',Name).

get_xtttable_name(_,Name):-
	gensym(xtt,Name).

get_xtt_name(BpmnNameGenerator,_GeneratorGoal,Name):-
	(is_bpmn_atts_names),
	get_xtt_name_generator_to_name(BpmnNameGenerator,Name),
	BpmnNameGenerator,!.

get_xtt_name(_,GeneratorGoal,Name):-
	get_xtt_name_generator_to_name(GeneratorGoal,Name),
	GeneratorGoal.

get_xtt_name_generator_to_name(Generator,Name):-
	Generator =.. [_,Name],!;
	Generator =.. [_,_,Name],!;
	Generator =.. [_,_,_,Name]. % if needed may be extended to more args generators





%%%%%%%%	helper predicates	%%%%%%%%


member_all([],_):-!.
member_all([M|Members],L):-
	select(M,L,R),!,
	member_all(Members,R).

list_is_empty([]).
remove_empty(Lists,NoEmptyLists):-
	exclude(list_is_empty,Lists,NoEmptyLists).

list_extract([],_,[]):-!.
list_extract([L|List],Extractor,Extr):-
	functor(F,Extractor,2),
	arg(1,F,L),
	arg(2,F,E),
	(
	 (F,!,Extr=[E|Es], list_extract(List,Extractor,Es))

	 ;                 list_extract(List,Extractor,Extr)
	).

list_extract_flatten(List,Extractor,ExtractedFlatten):-
	list_extract(List,Extractor,Extracted),
	flatten(Extracted,ExtractedFlatten).

write_listofpreds(Stream,Writter,ListOfPreds):-
	functor(Wr,Writter,2),arg(1,Wr,Stream),
	forall(member(P,ListOfPreds),
	       (copy_term(Wr,Wrr),arg(2,Wrr,P),Wrr)).
write_listofpreds(Stream,ListOfPreds):-
	write_listofpreds(Stream,write_to_stream,ListOfPreds).


write_to_stream(Stream,Term):-
	mywrite(Stream,Term),
	swritef(S,'.\n',[]),write(Stream,S).

stream_writeln(Stream,Term):-
	mywrite(Stream,Term),
	stream_nl(Stream).
stream_nl(Stream):-
	swritef(S,'\n',[]),write(Stream,S).

mywrite(Stream,T):-
	writeq(Stream,T).

%%	atom_to_codes_qout(Atom,Codes)
%
%	Preserves atoms quotations.

atom_to_codes_qout(Atom,Codes):-
	myatom_to_codes(Atom,Codes,[]).
atom_to_codes_qout(Atom,Codes,Tail):-
	with_output_to_chars(writeq(Stream,Atom),Stream,Codes,Tail).

atom_to_codes_qout_tailspace(Atom,Codes,Tail):-
	atom_to_codes_qout(Atom,Codes,[32|Tail]).





%%%%%%%%	DCG		%%%%%%%%

white_non_space(C):-
	white_char1('\r',C),!;
	white_char1('\n',C),!;
	white_char1('\t',C).

space_char(C):-
	white_char1(' ',C).

white_char(C):-
	white_non_space(C),!;
	space_char(C).

white_char1(Atom,Char):-
	atom_codes(Atom,Char).

non_whites([C|Cs]) -->
	[C],
	{\+ white_char([C])},
	non_whites(Cs).
non_whites([]) -->
	[].

whites([C|Cs]) -->
	[C],
	{white_char([C])},
	whites(Cs).
whites([]) -->
	[].

whites_non_spaces([C|Cs]) -->
	[C],
	{white_non_space([C])},
	whites_non_spaces(Cs).
whites_non_spaces([]) -->
	[].

non_whites_non_spaces([C|Cs]) -->
	[C],
	{\+ white_non_space([C])},
	non_whites_non_spaces(Cs).
non_whites_non_spaces([]) -->
	[].


remove_white_non_spaces(In,Out):-
	atom_codes(In,InCodes),
	remove_white_non_spaces1(InCodes,[],OutCodes),!,
	atom_codes(Out,OutCodes).

remove_white_non_spaces1([],Out,Out):-!.
remove_white_non_spaces1(In,Out0,Out):-
	phrase(whites_non_spaces(_W),In,Rest1),
	phrase(non_whites_non_spaces(Out1),Rest1,Rest2),

	ensure_space(Out0,Out1,Out01,Out11),
	append(Out01,Out11,Out2),

	remove_white_non_spaces1(Rest2,Out2,Out).


ensure_space([],[],[],[]):-!.
ensure_space(L,[],L,[]):-!.
ensure_space([],L,[],L):-!.
ensure_space(L1,L2,L11,L21):-
	(last(L1,N1),space_char([N1])
	;
	 L2=[N2|_],space_char([N2])),!,
	L11=L1,L21=L2.
ensure_space(L1,L2,L1,[S|L2]):-
	space_char([S]).


%%%
select_all(Elems,List,Selected,Rest):-
	select_all1(Elems,List,Selected,Rest).

select_all1([],List,[],List):-!.
select_all1(_,[],[],[]):-!.
select_all1([E|Es],List,[E|Ts],Rest):-
	select(E,List,R1),!,
	select_all1(Es,R1,Ts,Rest).

select_all1(_,List,[],List).


insert(X):-
	\+ X,!,
	assert(X).
insert(_).

delete(X):-retractall(X).
